#  tests for python-libarchive-c-2.9-py39h6ffa863_0 (this is a generated file);
print('===== testing package: python-libarchive-c-2.9-py39h6ffa863_0 =====');
print('running run_test.py');
#  --- run_test.py (begin) ---
import libarchive
from os.path import join

with libarchive.file_reader(join('test', 'hello_world.xar')) as archive:
    for entry in archive:
        print(entry.name)

#  --- run_test.py (end) ---

print('===== python-libarchive-c-2.9-py39h6ffa863_0 OK =====');
print("import: 'libarchive'")
import libarchive

